# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-10-22 11:21:32 +0900 (Wed, 22 Oct 2003) $
#   $LastChangedBy: sugi $
#

require 'vfs/common'
require 'vfs/file'
require 'vfs/dir'

Object.constants.member?("FileBuiltIn") or
  FileBuiltIn = File.clone
  
Object.constants.member?("DirBuiltIn") or
  DirBuiltIn = Dir.clone


module VFS
  module CompatLayer

    class Stat
      include Comparable

      def initialize(path)
        @path = path
      end

      def <=>(other)
        VFS::File.new(@path).mtime <=> VFS::File.new(other).mtime
      end

      class_eval [
       :uid, :mtime, :writable?, :zero?, :blockdev?, :dev,
       :grpowned?, :rdev_major, :executable_real?, :setgid?, :nlink,
       :size, :atime, :readable_real?, :socket?, :blocks,
       :directory?, :owned?, :rdev, :executable?, :setuid?, :mode,
       :ftype, :readable?, :symlink?, :blksize, :size?, :gid, :ctime,
       :writable_real?, :chardev?, :ino, :pipe?, :rdev_minor, :file?,
       :sticky?
      ].map { |m| %{
        def #{m}
          VFS::File.new(@path).#{m}
        end
      } }.join

    end # Stat

    class File < VFS::File
      class << self
        def new(path, *args, &block)
          VFS::File.new(path).open(*args, &block)
        end

        def umask(*mask)
          FileBuiltIn.umask(*mask)
        end

        def utime(n_atime, n_mtime, *files)
          files.each { |f|
            VFS::File.new(f).utime(n_atime, n_mtime)
          }
        end
        
        def chmod(mode, *files)
          files.each { |f|
            VFS::File.new(f).chmod(mode)
          }
        end
        
        def lchmod
          files.each { |f|
            VFS::File.new(f).lchmod(mode)
          }
        end
        
        def chown(user, group, *files)
          files.each { |f|
            VFS::File.new(f).chown(user, group)
          }
        end

        def lchown(user, group, *files)
          files.each { |f|
            VFS::File.new(f).chown(user, group)
          }
        end

        def unlink(*files)
          files.each { |f|
            VFS::File.new(f).unlink
          }
        end 
        alias :delete :unlink

        def stat(path)
          if VFS.scheme(path) == "file"
            FileBuiltIn.stat(path)
          else
            VFS::CompatLayer::Stat.new(path)
          end
        end

#        undef_method :mtime=, :atime=, :uid, :dev, :rdev_major, :nlink,
#          :blocks, :rdev, :mode, :blksize, :gid, :ino, :rdev_minor

      end # class << self(CompatLayer::File)
    end  # CompatLayer::File

    class Dir < VFS::Dir
      def chroot(path)
        DirBuiltIn.chroot(path)
      end

      def chdir(*args, &block)
        DirBuiltIn.chdir(*args, &block)
      end
    end # Dir

    extend VFS::MethodGenerator

    @@permanently  = false
    @@current      = :BuiltIn
    @@file_m_diff = []
    @@dir_m_diff  = []
    
    class_attr_accessor :permanently, :current

    class << self
      def allow_colon3_syntax?
        return false if ::VERSION.split(/\./)[0..2].join(".").to_f < 1.8
        return false if ::VERSION == "1.8.0" &&
                        ::RUBY_RELEASE_DATE.gsub(/-/, '').to_i < 20030814
        true
      end
      def allow_methods_with_flag?
        ::VERSION.split(/\./)[0..1].join(".").to_f < 1.7 ? false : true
      end

      def log(msg, *level)
	VFS::Log.log("VFS(CompatLayer): " + msg, *level)
      end

      def override
	unless allow_methods_with_flag?
	  return false
	end
	if @@current != :BuiltIn
	  log("built-in class was already overrided? skip override.", Log::DEBUG)
	  return true
	end
	log("overriding built-in class.", Log::DEBUG)
	override_kernel_open
	@@file_m_diff = gen_class_method_delegator(::File,
                          [::VFS::CompatLayer::File, ::VFS::File],
                          @@file_m_diff)
	@@dir_m_diff  = gen_class_method_delegator(::Dir,
                          [::VFS::CompatLayer::Dir, ::VFS::Dir],
                          @@dir_m_diff)
	@@current = :VFS
      end

      def restore
	unless allow_methods_with_flag?
	  return false
	end
	if @@current != :VFS
	  log("built-in class was not overrided? skip restore.", Log::DEBUG)
	  return true
	end
	if @@permanently
	  log("permanent flug is ture. skip restore.", Log::DEBUG)
	  return true
	end
	restore_kernel_open
	@@file_m_diff = gen_class_method_delegator(::File, ::FileBuiltIn,
                                                   @@file_m_diff)
	@@dir_m_diff  = gen_class_method_delegator(::Dir,  ::DirBuiltIn,
                                                   @@dir_m_diff)
	@@current = :BuiltIn
      end

      def override_kernel_open
	log("overriding kernel open. go eval...", Log::DEBUG)
        ::Kernel.module_eval %Q{
          begin
            # why I can not use respond_to? :open_bulitin ???
            method :open_builtin
            # nothing to do
          rescue NameError
            alias_method :open_builtin, :open
            def open(*args, &block)
              VFS::File.open(*args, &block)
            end
          end # rescue NameError
          begin
            # why I can not use respond_to? :open_bulitin ???
            method :open_builtin
            # nothing to do
          rescue NameError
            alias_method :open_builtin, :open
            def open(*args, &block)
              VFS::File.open(*args, &block)
            end
          end # rescue NameError
	}
      end

      def restore_kernel_open
	log("restoring kernel open. go eval...", Log::DEBUG)
        ::Kernel.module_eval %Q{
	    begin
	      # why I can not use respond_to? :open_bulitin ???
	      method :open_builtin
	      remove_method :open
	      alias_method  :open, :open_builtin
	      remove_method :open_builtin
	    rescue NameError
	      # nothing to do
	    end
	}
      end

      # return added methods
      def gen_class_method_delegator(target_class, class_new, del_methods=[])
        class_new = [class_new] unless Array === class_new
	log("gen class methods for #{target_class} to #{class_new.join(', ')}.",
            Log::DEBUG)
        new_methods = class_new.map{|c| c.methods(false) }.flatten.uniq
        old_methods = target_class.methods(false)
        rem_methods = old_methods - new_methods - del_methods

        eval_str    = "class << self;" 
        eval_str   << %{
          remove_method #{(old_methods & new_methods).map{|m| ":#{m}"}.join(",")}
        }

        eval_str << new_methods.map { |m|
	  if m.to_s =~ /=$/
            %{
              def #{m}(*args)
                ::#{class_new[0].to_s}.#{m} *args
              end
            }
          else
            %{
              def #{m}(*args, &block)
                ::#{class_new[0].to_s}.#{m}(*args, &block)
              end
            }
          end
        }.join

        if del_methods.length > 0
          eval_str << "remove_method " <<
            del_methods.map{|m| ":#{m}"}.join(",")  << "\n"
          VFS::Log.log("VFS(CompatLayer): following methods was deleted." +
		       " (#{del_methods.join(', ')})", VFS::Log::DEBUG)
        end
        if rem_methods.length > 0
          VFS::Log.log("VFS(CompatLayer): following methods was remaind." +
		       " (#{rem_methods.join(', ')})", VFS::Log::INFO)
        end

        eval_str << ";end"
        target_class.class_eval eval_str
        new_methods - old_methods
      end # gen_class_method_delegator

    end # class << self
  end # CompatLayer
end # VFS


unless VFS::CompatLayer.allow_methods_with_flag?
  # simple override is only for ruby 1.6 and older.
  # use VFS::ConpatLayer above for 1.8 and later.
  # but eval may be very slow...
  #
  module Kernel
    def open(*args, &block)
      VFS::File.open(*args, &block)
    end
  end
  
  class File
    class << self
      method_def = ""
      (VFS::File.instance_methods(false) + [:new]).each {|m|
	m.to_s =~ /=$/ and next
	begin
	  remove_method m
	rescue NameError
	end
	class_eval %Q{
          def #{m}(*args, &block)
	    VFS::CompatLayer::File.#{m}(*args, &block)
	  end
	}
      }
    end
  end
  
  class Dir
    class << self
      (VFS::Dir.instance_methods(false) +[:new]).each {|m|
	m.to_s =~ /=$/ and next
	begin
	  remove_method m
	rescue NameError
	end
	class_eval %Q{
          def #{m}(*args, &block)
	    VFS::CompatLayer::Dir.#{m}(*args, &block)
	  end
	}
      }
    end
  end
end
